# featureBased.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com> 

"""
This file contains code that implements feature based submodular selection
algorithms.
"""

import numpy

from .base import SubmodularSelection

from tqdm import tqdm

from numba import njit
from numba import prange

dtypes = 'int64(float64[:,:], float64[:], float64[:], float64[:], int8[:])'

@njit(dtypes, nogil=True, parallel=True)
def select_sqrt_next(X, gains, current_values, current_concave_values, mask):
	for idx in prange(X.shape[0]):
		if mask[idx] == 1:
			continue

		a = numpy.sqrt(current_values + X[idx])
		gains[idx] = (a - current_concave_values).sum()

	return numpy.argmax(gains)

@njit(dtypes, nogil=True, parallel=True)
def select_log_next(X, gains, current_values, current_concave_values, mask):
	for idx in prange(X.shape[0]):
		if mask[idx] == 1:
			continue

		a = numpy.log(current_values + X[idx] + 1)
		gains[idx] = (a - current_concave_values).sum()

	return numpy.argmax(gains)

@njit(dtypes, nogil=True, parallel=True)
def select_inv_next(X, gains, current_values, current_concave_values, mask):
	for idx in prange(X.shape[0]):
		if mask[idx] == 1:
			continue

		a = (current_values + X[idx]) / (1. + current_values + X[idx])
		gains[idx] = (a - current_concave_values).sum()

	return numpy.argmax(gains)

@njit(dtypes, nogil=True, parallel=True)
def select_min_next(X, gains, current_values, current_concave_values, mask):
	for idx in prange(X.shape[0]):
		if mask[idx] == 1:
			continue

		a = numpy.fmin(current_values + X[idx], numpy.ones(X.shape[1]))
		gains[idx] = (a - current_concave_values).sum()

	return numpy.argmax(gains)

def select_custom_next(X, gains, current_values, current_concave_values, mask, 
	concave_func):
	best_gain = 0.
	best_idx = -1

	for idx in range(X.shape[0]):
		if mask[idx] == 1:
			continue

		a = concave_func(current_values + X[idx])
		gains[idx] = (a - current_concave_values).sum()

		if gains[idx] > best_gain:
			best_gain = gain
			best_idx = idx

	return best_idx

class FeatureBasedSelection(SubmodularSelection):
	"""A feature based submodular selection algorithm.

	NOTE: All values in your data must be positive for this selection to work.

	This function will use a feature based submodular selection algorithm to
	identify a representative subset of the data. The feature based functions
	use the values of the features themselves, rather than a transformation of
	the values, in order to select a diverse subset. The goal of this approach
	is to find a representative subset that sees each ~feature~ at a certain
	saturation across the selected points, rather than trying to uniformly
	sample the space.

	This implementation uses the lazy greedy algorithm so that multiple passes
	over the whole data set are not required each time a sample is selected. The
	benefit of this effect is not typically seen until many (over 100) passes are
	seen of the data set.

	Parameters
	----------
	n_samples : int
		The number of samples to return.

	concave_func : str or callable
		The type of concave function to apply to the feature values. You can
		pass in your own function to apply. Otherwise must be one of the
		following:

			'log' : log(1 + X)
			'sqrt' : sqrt(X)
			'min' : min(X, 1)
			'inverse' : X / (1 + X)

	verbose : bool
		Whether to print output during the selection process.

	Attributes
	----------
	pq : PriorityQueue
		The priority queue used to implement the lazy greedy algorithm.

	n_samples : int
		The number of samples to select.

	concave_func : callable
		A concave function for transforming feature values, often referred to as
		phi in the literature.

	ranking : numpy.array int
		The selected samples in the order of their gain with the first number in
		the ranking corresponding to the index of the first sample that was
		selected by the greedy procedure.
	"""

	def __init__(self, n_samples, concave_func='sqrt', n_greedy_samples=3, 
		verbose=False):
		self.concave_func_name = concave_func

		if concave_func == 'log':
			self.concave_func = lambda X: numpy.log(X + 1)
		elif concave_func == 'sqrt':
			self.concave_func = lambda X: numpy.sqrt(X)
		elif concave_func == 'min':
			self.concave_func = lambda X: numpy.fmin(X, numpy.ones_like(X))
		elif concave_func == 'inverse':
			self.concave_func = lambda X: X / (1. + X)
		elif callable(concave_func):
			self.concave_func = concave_func
		else:
			raise KeyError("Must be one of 'log', 'sqrt', 'min', 'inverse', or a custom function.")

		super(FeatureBasedSelection, self).__init__(n_samples, n_greedy_samples, verbose)

	def fit(self, X, y=None):
		"""Perform selection and return the subset of the data set.

		This method will take in a full data set and return the selected subset
		according to the feature based function. The data will be returned in
		the order that it was selected, with the first row corresponding to
		the best first selection, the second row corresponding to the second
		best selection, etc.

		Parameters
		----------
		X : list or numpy.ndarray, shape=(n, d)
			The data set to transform. Must be numeric.

		y : list or numpy.ndarray, shape=(n,), optional
			The labels to transform. If passed in this function will return
			both the data and th corresponding labels for the rows that have
			been selected.

		Returns
		-------
		self : FeatureBasedSelection
			The fit step returns itself.
		"""

		return super(FeatureBasedSelection, self).fit(X, y)

	def _greedy_select(self, X):
		"""Select elements in a naive greedy manner."""

		concave_funcs = {
			'sqrt': select_sqrt_next,
			'log': select_log_next,
			'inverse': select_inv_next,
			'min': select_min_next
		}

		concave_func = concave_funcs.get(self.concave_func_name, 
			select_custom_next)

		for i in range(self.n_greedy_samples):
			gains = numpy.zeros(X.shape[0], dtype='float64')
			
			if self.concave_func_name in concave_funcs:
				best_idx = concave_func(X, gains, self.current_values,
					self.current_concave_values, self.mask)
			else:
				best_idx = select_custom_next(X, gains, current_values, 
					current_concave_values, mask, self.concave_func)

			self.ranking.append(best_idx)
			self.mask[best_idx] = True
			self.current_values += X[best_idx]
			self.current_concave_values = self.concave_func(self.current_values)

			if self.verbose == True:
				self.pbar.update(1)

		return gains

	def _lazy_greedy_select(self, X):
		"""Select elements from a dense matrix in a lazy greedy manner."""

		for i in range(self.n_greedy_samples, self.n_samples):
			best_gain = 0.
			best_idx = None
			
			while True:
				prev_gain, idx = self.pq.pop()
				prev_gain = -prev_gain
				
				if best_gain >= prev_gain:
					self.pq.add(idx, -prev_gain)
					self.pq.remove(best_idx)
					break
				
				a = self.concave_func(self.current_values + X[idx])
				gain = (a - self.current_concave_values).sum()
				
				self.pq.add(idx, -gain)
				
				if gain > best_gain:
					best_gain = gain
					best_idx = idx

			self.ranking.append(best_idx)
			self.mask[best_idx] = True
			self.current_values += X[best_idx]
			self.current_concave_values = self.concave_func(self.current_values)

			if self.verbose == True:
				self.pbar.update(1)



# facilityLocation.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

"""
This code implements facility location functions.
"""

import numpy

from .base import SubmodularSelection

from tqdm import tqdm

from numba import njit
from numba import prange

from scipy.sparse import csr_matrix

dtypes = 'int64(float64[:,:], float64[:], float64[:], int8[:])'
sdtypes = 'int64(float64[:], int32[:], int32[:], float64[:], float64[:], int8[:])'

@njit(dtypes, nogil=True, parallel=True)
def select_next(X, gains, current_values, mask):
	for idx in prange(X.shape[0]):
		if mask[idx] == 1:
			continue

		a = numpy.maximum(X[idx], current_values)
		gains[idx] = (a - current_values).sum()

	return numpy.argmax(gains)

@njit(sdtypes, nogil=True, parallel=True)
def select_next_sparse(X_data, X_indices, X_indptr, gains, current_values, mask):
	for idx in range(X_indptr.shape[0] - 1):
		if mask[idx] == 1:
			continue

		start = X_indptr[idx]
		end = X_indptr[idx+1]

		for i in range(start, end):
			j = X_indices[i]

			if X_data[i] > current_values[j]:
				gains[idx] += X_data[i] - current_values[j]

	return numpy.argmax(gains)


class FacilityLocationSelection(SubmodularSelection):
	"""A facility location submodular selection algorithm.

	NOTE: All ~pairwise~ values in your data must be positive for this 
	selection to work.

	This function uses a facility location based submodular selection algorithm
	to identify a representative subset of the data. This feature based function
	works on pairwise relationships between each of the samples. This can be
	the correlation, a dot product, or any other such function where a higher
	value corresponds to a higher similarity and a lower value corresponds to
	a lower similarity.

	This implementation allows users to pass in either their own symmetric
	square matrix of similarity values, or a data matrix as normal and a function
	that calculates these pairwise values.

	Parameters
	----------
	n_samples : int
		The number of samples to return.

	pairwise_func : str or callable
		The method for converting a data matrix into a square symmetric matrix
		of pairwise similarities. If a string, can be any of the following:

			'euclidean' : The negative euclidean distance
			'corr' : The squared correlation matrix
			'cosine' : The normalized dot product of the matrix
			'precomputed' : User passes in a NxN matrix of distances themselves

	Attributes
	----------
	n_samples : int
		The number of samples to select.

	pairwise_func : callable
		A function that takes in a data matrix and converts it to a square
		symmetric matrix.

	ranking : numpy.array int
		The selected samples in the order of their gain.
	"""

	def __init__(self, n_samples=10, pairwise_func='euclidean', n_greedy_samples=1, 
		verbose=False):
		self.pairwise_func_name = pairwise_func
		
		norm = lambda x: numpy.sqrt((x*x).sum(axis=1)).reshape(x.shape[0], 1)
		norm2 = lambda x: (x*x).sum(axis=1).reshape(x.shape[0], 1)

		if pairwise_func == 'corr':
			self.pairwise_func = lambda X: numpy.corrcoef(X, rowvar=True) ** 2.
		elif pairwise_func == 'cosine':
			self.pairwise_func = lambda X: numpy.abs(numpy.dot(X, X.T) / (norm(X).dot(norm(X).T)))
		elif pairwise_func == 'euclidean':
			self.pairwise_func = lambda X: -((-2 * numpy.dot(X, X.T) + norm2(X)).T + norm2(X))
		elif pairwise_func == 'precomputed':
			self.pairwise_func = pairwise_func
		elif callable(pairwise_func):
			self.pairwise_func = pairwise_func
		else:
			raise KeyError("Must be one of 'euclidean', 'corr', 'cosine', 'precomputed'" \
				" or a custom function.")

		super(FacilityLocationSelection, self).__init__(n_samples, 
			n_greedy_samples, verbose)

	def fit(self, X, y=None):
		"""Perform selection and return the subset of the data set.

		This method will take in a full data set and return the selected subset
		according to the facility location function. The data will be returned in
		the order that it was selected, with the first row corresponding to
		the best first selection, the second row corresponding to the second
		best selection, etc.

		Parameters
		----------
		X : list or numpy.ndarray, shape=(n, d)
			The data set to transform. Must be numeric.

		y : list or numpy.ndarray, shape=(n,), optional
			The labels to transform. If passed in this function will return
			both the data and th corresponding labels for the rows that have
			been selected.

		Returns
		-------
		self : FacilityLocationSelection
			The fit step returns itself.
		"""

		f = self.pairwise_func

		if isinstance(X, csr_matrix) and f != "precomputed":
			raise ValueError("Must passed in a precomputed sparse " \
				"similarity  matrix or a dense feature matrix.")
		if f == 'precomputed' and X.shape[0] != X.shape[1]:
			raise ValueError("Precomputed similarity matrices " \
				"must be square and symmetric.")

		if self.verbose == True:
			self.pbar = tqdm(total=self.n_samples)
			self.pbar.update(1)

		self.sparse = isinstance(X, csr_matrix)

		if self.pairwise_func == 'precomputed':
			X_pairwise = X
		else:
			X = numpy.array(X, dtype='float64')
			X_pairwise = self.pairwise_func(X)

		return super(FacilityLocationSelection, self).fit(X_pairwise, y)

	def _greedy_select(self, X_pairwise):
		"""Select elements in a naive greedy manner."""

		for i in range(self.n_greedy_samples):
			gains = numpy.zeros(X_pairwise.shape[0], dtype='float64')
		
			if not self.sparse:
				best_idx = select_next(X_pairwise, gains, self.current_values,
					self.mask)
				self.current_values = numpy.maximum(X_pairwise[best_idx], 
					self.current_values)
			else:
				best_idx = select_next_sparse(X_pairwise.data,
					X_pairwise.indices, X_pairwise.indptr, gains,
					self.current_values, self.mask)
				self.current_values = numpy.maximum(
					X_pairwise[best_idx].toarray()[0], self.current_values)

			self.ranking.append(best_idx)
			self.mask[best_idx] = 1

			if self.verbose == True:
				self.pbar.update(1)

		return gains

	def _lazy_greedy_select(self, X_pairwise):
		"""Select elements from a dense matrix in a lazy greedy manner."""

		for i in range(self.n_greedy_samples, self.n_samples):
			best_gain = 0.
			best_idx = None
			
			while True:
				prev_gain, idx = self.pq.pop()
				prev_gain = -prev_gain
				
				if best_gain >= prev_gain:
					self.pq.add(idx, -prev_gain)
					self.pq.remove(best_idx)
					break
				
				if not self.sparse:
					a = numpy.maximum(X_pairwise[:, idx], 
						self.current_values)

					gain = (a - self.current_values).sum()
				else:
					gain = 0.
					start = X_pairwise.indptr[idx]
					end = X_pairwise.indptr[idx+1]

					for k in range(start, end):
						j = X_pairwise.indices[k]

						if X_pairwise.data[k] > self.current_values[j]:
							gain += (X_pairwise.data[k] - 
								self.current_values[j])

				self.pq.add(idx, -gain)
				
				if gain > best_gain:
					best_gain = gain
					best_idx = idx

			self.ranking.append(best_idx)
			self.mask[best_idx] = True

			if not self.sparse:
				self.current_values = numpy.maximum(X_pairwise[best_idx],
					self.current_values)
			else:
				start = X_pairwise.indptr[best_idx]
				end = X_pairwise.indptr[best_idx+1]

				for k in range(start, end):
					j = X_pairwise.indices[k]
					self.current_values[j] = max(X_pairwise.data[k], 
						self.current_values[j])

			if self.verbose == True:
				self.pbar.update(1)




# base.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com> 

"""
This file contains code that implements the core of the submodular selection
algorithms.
"""

import numpy

from .utils import PriorityQueue

from tqdm import tqdm
from scipy.sparse import csr_matrix

class SubmodularSelection(object):
	"""The base selection object.

	This object defines the structures that all submodular selection algorithms
	should follow. All algorithms will have the same public methods and the
	same attributes.

	Parameters
	----------
	n_samples : int
		The number of samples to return.

	concave_func : str or callable
		The type of concave function to apply to the feature values. You can
		pass in your own function to apply. Otherwise must be one of the
		following:

			'log' : log(1 + X)
			'sqrt' : sqrt(X)
			'min' : min(X, 1)
			'inverse' : 1 / (1 + X)

	verbose : bool
		Whether to print output during the selection process.

	Attributes
	----------
	pq : PriorityQueue
		The priority queue used to implement the lazy greedy algorithm.

	n_samples : int
		The number of samples to select.

	ranking : numpy.array int
		The selected samples in the order of their gain with the first number in
		the ranking corresponding to the index of the first sample that was
		selected by the greedy procedure.
	"""

	def __init__(self, n_samples, n_greedy_samples=1, verbose=False):
		self.pq = PriorityQueue()

		if type(n_samples) != int:
			raise ValueError("n_samples must be a positive integer.")
		if n_samples < 1:
			raise ValueError("n_samples must be a positive integer.")

		if type(n_greedy_samples) != int:
			raise ValueError("n_greedy_samples must be a positive integer")
		if n_greedy_samples < 1:
			raise ValueError("n_greedy_samples must be a positive integer.")
		if n_greedy_samples > n_samples:
			raise ValueError("n_samples must be larger than n_greedy_samples")

		if verbose not in (True, False):
			raise ValueError("verbosity must be True or False")

		self.n_samples = n_samples
		self.n_greedy_samples = n_greedy_samples
		self.verbose = verbose
		self.ranking = None
		self.sparse = None
	
	def fit(self, X, y=None):
		"""Fit a ranking to the data set of the top n_sample elements.

		This method will take in the data set (and optionally a label) set
		and fit a ranking to it. This runs the greedy select procedure n_samples
		times, selecting at each iteration the next best element. The ranking
		is then stored for future use.

		Parameters
		----------
		X : list or numpy.ndarray, shape=(n, d)
			The data set to transform. Must be numeric.

		y : list or numpy.ndarray, shape=(n,), optional
			The labels to transform. If passed in this function will return
			both the data and th corresponding labels for the rows that have
			been selected.

		Returns
		-------
		self : SubmodularSelection
		"""

		allowed_dtypes = list, numpy.ndarray, csr_matrix

		if not isinstance(X, allowed_dtypes):
			raise ValueError("X must be either a list of lists, a 2D numpy " \
				"array, or a scipy.sparse.csr_matrix.")
		if isinstance(X, numpy.ndarray) and len(X.shape) != 2:
			raise ValueError("X must have exactly two dimensions.")
		if numpy.min(X) < 0.0 and numpy.max(X) > 0.:
			raise ValueError("X cannot contain negative values or must be entirely "\
				"negative values.")

		if self.verbose == True:
			self.pbar = tqdm(total=self.n_samples)

		self.current_values = numpy.zeros(X.shape[1], dtype='float64')
		self.current_values += numpy.min(X)
		self.current_concave_values = numpy.zeros(X.shape[1])

		self.mask = numpy.zeros(X.shape[0], dtype='int8')
		self.ranking = []

		# Select using the greedy algorithm first returning the gains from
		# the last round of selection.
		gains = self._greedy_select(X)

		# Populate the priority queue following greedy selection
		if self.n_greedy_samples < self.n_samples:
			for idx, gain in enumerate(gains):
				if self.mask[idx] != 1:
					self.pq.add(idx, -gain) 

			# Now select remaining elements using the lazy greedy algorithm.
			self._lazy_greedy_select(X)

		if self.verbose == True:
			self.pbar.close()
		
		self.ranking = numpy.array(self.ranking)
		return self

	def _greedy_select(self, X):
		raise NotImplementedError

	def _lazy_greedy_select(self, X):
		raise NotImplementedError

	def transform(self, X, y=None):
		"""Transform the data set by selecting the top n_samples samples.

		This method will use the fit ranking to transform the data set by
		returning only the top n_samples elements as defined by the
		ranking. The data will be returned in the order that it was selected
		during the selection process.

		Parameters
		----------
		X : list or numpy.ndarray, shape=(n, d)
			The data set to transform. Must be numeric.

		y : list or numpy.ndarray, shape=(n,), optional
			The labels to transform. If passed in this function will return
			both the data and th corresponding labels for the rows that have
			been selected.

		Returns
		-------
		X_subset : numpy.ndarray, shape=(n_samples, d)
			A subset of the data such that n_samples < n and n_samples is the
			integer provided at initialization.

		y_subset : numpy.ndarray, shape=(n_samples,)
			The labels that match with the indices of the samples if y is
			passed in. Only returned if passed in.
		"""

		if y is not None:
			return X[self.ranking], y[self.ranking]
		return X[self.ranking]

	def fit_transform(self, X, y=None):
		"""Perform selection and return the subset of the data set.

		This method will take in a full data set and return the selected subset
		according to the feature based function. The data will be returned in
		the order that it was selected, with the first row corresponding to
		the best first selection, the second row corresponding to the second
		best selection, etc.

		Parameters
		----------
		X : list or numpy.ndarray, shape=(n, d)
			The data set to transform. Must be numeric.

		y : list or numpy.ndarray, shape=(n,), optional
			The labels to transform. If passed in this function will return
			both the data and th corresponding labels for the rows that have
			been selected.

		Returns
		-------
		X_subset : numpy.ndarray, shape=(n_samples, d)
			A subset of the data such that n_samples < n and n_samples is the
			integer provided at initialization.

		y_subset : numpy.ndarray, shape=(n_samples,)
			The labels that match with the indices of the samples if y is
			passed in. Only returned if passed in.
		"""

		return self.fit(X, y).transform(X, y)



# utils.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

"""
This code contains utility functions to support the main functionality of
the code.
"""

from heapq import heappush
from heapq import heappop

class PriorityQueue(object):
    """A priority queue implementation.

    This is an implemented of a priority queue using a heap implemtnation. It
    has O(log n) time for adding and popping elements but constant time
    deletion. It is based off of the implementation that is provided at

    https://docs.python.org/2/library/heapq.html

    This implementation assumes that the items with the highest priority
    are the ones with the lowest "weight" that gets passed in. If higher
    weights are supposed to correspond to higher priority, consider reversing
    the sign of the weight.

    Parameters
    ----------
    None

    Attributes
    ----------
    pq : list
        A list containing the heapified priority queue.

    lookup : dict
        A dictionary containing links to each of the elements, allowing
        constant time lookups and deletions.
    """

    def __init__(self):
        self.pq = []
        self.lookup = {}
    
    def add(self, item, weight):
        """Add an element to the priority queue. Runtime is O(log n).

        This adds in an element to the priority queue. If that element is
        already there it will automatically delete the previous instance
        and replace it with the new instance.

        Parameters
        ----------
        item : object
            The object to be encoded.

        weight : double
            The priority of the item. The lower the weight the higher the
            priority when items get dequeued. If a higher weight is supposed
            to correspond to a higher priority, consider reversing the sign of
            the weight.

        Returns
        -------
        None
        """



        if item in self.lookup:
            self.remove(item)
        
        entry = [weight, item]
        self.lookup[item] = entry
        heappush(self.pq, entry)
    
    def remove(self, item):
        """Remove an element from the queue.

        This is not popping the highest priority item, rather it will remove
        an element that is present in the queue. If one attempts to remove an
        item that is not present in the queue the function will error.
        
        Parameters
        ----------
        item : object
            The object to be removed from the queue.

        Returns
        -------
        None
        """


        entry = self.lookup.pop(item)
        entry[-1] = "DELETED"
    
    def pop(self):
        """Pop the highest priority element from the queue. Runtime is O(log n).

        This will remove the highest priority element from the queue. If there
        are no elements left in the queue it will raise an error.

        Parameters
        ----------
        None

        Returns
        -------
        weight : double
            The weight of the element as passed in in the `add` method

        item : object
            The item that was passed in during the `add` method
        """

        while self.pq:
            weight, item = heappop(self.pq)
            if item != "DELETED":
                del self.lookup[item]
                return weight, item
        
        raise KeyError("No elements left in the priority queue.")
